@@ -122,6 +122,17 @@ public:
     timer_queues_.clear();
   } 
 
+  // Initialise the task, but only if the reactor is not in its own thread.
+  void init_task()
+  {
+    if (!Own_Thread)
+    {
+      typedef task_io_service<dev_poll_reactor<Own_Thread> >
+        task_io_service_type;
+      use_service<task_io_service_type>(this->get_io_service()).init_task();
+    }
+  }
+
   // Register a socket with the reactor. Returns 0 on success, system error
   // code on failure.
   int register_descriptor(socket_type, per_descriptor_data&)
@@ -422,14 +433,15 @@ private:
         else
           more_writes = write_op_queue_.has_operation(descriptor);
 
-        if ((events[i].events == POLLHUP)
-            && !more_except && !more_reads && !more_writes)
+        if ((events[i].events & (POLLERR | POLLHUP)) != 0
+              && (events[i].events & ~(POLLERR | POLLHUP)) == 0
+              && !more_except && !more_reads && !more_writes)
         {
-          // If we have only an POLLHUP event and no operations associated
-          // with the descriptor then we need to delete the descriptor from
-          // /dev/poll. The poll operation might produce POLLHUP events even
-          // if they are not specifically requested, so if we do not remove the
-          // descriptor we can end up in a tight polling loop.
+          // If we have an event and no operations associated with the
+          // descriptor then we need to delete the descriptor from /dev/poll.
+          // The poll operation can produce POLLHUP or POLLERR events when there
+          // is no operation pending, so if we do not remove the descriptor we
+          // can end up in a tight polling loop.
           ::pollfd ev = { 0 };
           ev.fd = descriptor;
           ev.events = POLLREMOVE;
