@@ -60,6 +60,13 @@ class kqueue_reactor
   : public asio::detail::service_base<kqueue_reactor<Own_Thread> >
 {
 public:
+  // Per-descriptor data.
+  struct per_descriptor_data
+  {
+    bool allow_speculative_read;
+    bool allow_speculative_write;
+  };
+
   // Constructor.
   kqueue_reactor(asio::io_service& io_service)
     : asio::detail::service_base<
@@ -124,19 +131,47 @@ public:
     timer_queues_.clear();
   }
 
+  // Initialise the task, but only if the reactor is not in its own thread.
+  void init_task()
+  {
+    if (!Own_Thread)
+    {
+      typedef task_io_service<kqueue_reactor<Own_Thread> > task_io_service_type;
+      use_service<task_io_service_type>(this->get_io_service()).init_task();
+    }
+  }
+
   // Register a socket with the reactor. Returns 0 on success, system error
   // code on failure.
-  int register_descriptor(socket_type)
+  int register_descriptor(socket_type, per_descriptor_data& descriptor_data)
   {
+    descriptor_data.allow_speculative_read = true;
+    descriptor_data.allow_speculative_write = true;
+
     return 0;
   }
 
   // Start a new read operation. The handler object will be invoked when the
   // given descriptor is ready to be read, or an error has occurred.
   template <typename Handler>
-  void start_read_op(socket_type descriptor, Handler handler,
+  void start_read_op(socket_type descriptor,
+      per_descriptor_data& descriptor_data, Handler handler,
       bool allow_speculative_read = true)
   {
+    if (allow_speculative_read && descriptor_data.allow_speculative_read)
+    {
+      asio::error_code ec;
+      std::size_t bytes_transferred = 0;
+      if (handler.perform(ec, bytes_transferred))
+      {
+        handler.complete(ec, bytes_transferred);
+        return;
+      }
+
+      // We only get one shot at a speculative read in this function.
+      allow_speculative_read = false;
+    }
+
     asio::detail::mutex::scoped_lock lock(mutex_);
 
     if (shutdown_)
@@ -145,8 +180,21 @@ public:
     if (!allow_speculative_read)
       need_kqueue_wait_ = true;
     else if (!read_op_queue_.has_operation(descriptor))
-      if (handler(asio::error_code()))
+    {
+      // Speculative reads are ok as there are no queued read operations.
+      descriptor_data.allow_speculative_read = true;
+
+      asio::error_code ec;
+      std::size_t bytes_transferred = 0;
+      if (handler.perform(ec, bytes_transferred))
+      {
+        handler.complete(ec, bytes_transferred);
         return;
+      }
+    }
+
+    // Speculative reads are not ok as there will be queued read operations.
+    descriptor_data.allow_speculative_read = false;
 
     if (read_op_queue_.enqueue_operation(descriptor, handler))
     {
@@ -156,7 +204,7 @@ public:
       {
         asio::error_code ec(errno,
             asio::error::get_system_category());
-        read_op_queue_.dispatch_all_operations(descriptor, ec);
+        read_op_queue_.perform_all_operations(descriptor, ec);
       }
     }
   }
@@ -164,9 +212,24 @@ public:
   // Start a new write operation. The handler object will be invoked when the
   // given descriptor is ready to be written, or an error has occurred.
   template <typename Handler>
-  void start_write_op(socket_type descriptor, Handler handler,
+  void start_write_op(socket_type descriptor,
+      per_descriptor_data& descriptor_data, Handler handler,
       bool allow_speculative_write = true)
   {
+    if (allow_speculative_write && descriptor_data.allow_speculative_write)
+    {
+      asio::error_code ec;
+      std::size_t bytes_transferred = 0;
+      if (handler.perform(ec, bytes_transferred))
+      {
+        handler.complete(ec, bytes_transferred);
+        return;
+      }
+
+      // We only get one shot at a speculative write in this function.
+      allow_speculative_write = false;
+    }
+
     asio::detail::mutex::scoped_lock lock(mutex_);
 
     if (shutdown_)
@@ -175,8 +238,21 @@ public:
     if (!allow_speculative_write)
       need_kqueue_wait_ = true;
     else if (!write_op_queue_.has_operation(descriptor))
-      if (handler(asio::error_code()))
+    {
+      // Speculative writes are ok as there are no queued write operations.
+      descriptor_data.allow_speculative_write = true;
+
+      asio::error_code ec;
+      std::size_t bytes_transferred = 0;
+      if (handler.perform(ec, bytes_transferred))
+      {
+        handler.complete(ec, bytes_transferred);
         return;
+      }
+    }
+
+    // Speculative writes are not ok as there will be queued write operations.
+    descriptor_data.allow_speculative_write = false;
 
     if (write_op_queue_.enqueue_operation(descriptor, handler))
     {
@@ -186,7 +262,7 @@ public:
       {
         asio::error_code ec(errno,
             asio::error::get_system_category());
-        write_op_queue_.dispatch_all_operations(descriptor, ec);
+        write_op_queue_.perform_all_operations(descriptor, ec);
       }
     }
   }
@@ -194,7 +270,8 @@ public:
   // Start a new exception operation. The handler object will be invoked when
   // the given descriptor has exception information, or an error has occurred.
   template <typename Handler>
-  void start_except_op(socket_type descriptor, Handler handler)
+  void start_except_op(socket_type descriptor,
+      per_descriptor_data&, Handler handler)
   {
     asio::detail::mutex::scoped_lock lock(mutex_);
 
@@ -212,22 +289,25 @@ public:
       {
         asio::error_code ec(errno,
             asio::error::get_system_category());
-        except_op_queue_.dispatch_all_operations(descriptor, ec);
+        except_op_queue_.perform_all_operations(descriptor, ec);
       }
     }
   }
 
-  // Start new write and exception operations. The handler object will be
-  // invoked when the given descriptor is ready for writing or has exception
-  // information available, or an error has occurred.
+  // Start a new write operation. The handler object will be invoked when the
+  // given descriptor is ready to be written, or an error has occurred.
   template <typename Handler>
-  void start_write_and_except_ops(socket_type descriptor, Handler handler)
+  void start_connect_op(socket_type descriptor,
+      per_descriptor_data& descriptor_data, Handler handler)
   {
     asio::detail::mutex::scoped_lock lock(mutex_);
 
     if (shutdown_)
       return;
 
+    // Speculative writes are not ok as there will be queued write operations.
+    descriptor_data.allow_speculative_write = false;
+
     if (write_op_queue_.enqueue_operation(descriptor, handler))
     {
       struct kevent event;
@@ -236,23 +316,7 @@ public:
       {
         asio::error_code ec(errno,
             asio::error::get_system_category());
-        write_op_queue_.dispatch_all_operations(descriptor, ec);
-      }
-    }
-
-    if (except_op_queue_.enqueue_operation(descriptor, handler))
-    {
-      struct kevent event;
-      if (read_op_queue_.has_operation(descriptor))
-        EV_SET(&event, descriptor, EVFILT_READ, EV_ADD, 0, 0, 0);
-      else
-        EV_SET(&event, descriptor, EVFILT_READ, EV_ADD, EV_OOBAND, 0, 0);
-      if (::kevent(kqueue_fd_, &event, 1, 0, 0, 0) == -1)
-      {
-        asio::error_code ec(errno,
-            asio::error::get_system_category());
-        except_op_queue_.dispatch_all_operations(descriptor, ec);
-        write_op_queue_.dispatch_all_operations(descriptor, ec);
+        write_op_queue_.perform_all_operations(descriptor, ec);
       }
     }
   }
@@ -260,25 +324,15 @@ public:
   // Cancel all operations associated with the given descriptor. The
   // handlers associated with the descriptor will be invoked with the
   // operation_aborted error.
-  void cancel_ops(socket_type descriptor)
+  void cancel_ops(socket_type descriptor, per_descriptor_data&)
   {
     asio::detail::mutex::scoped_lock lock(mutex_);
     cancel_ops_unlocked(descriptor);
   }
 
-  // Enqueue cancellation of all operations associated with the given
-  // descriptor. The handlers associated with the descriptor will be invoked
-  // with the operation_aborted error. This function does not acquire the
-  // kqueue_reactor's mutex, and so should only be used from within a reactor
-  // handler.
-  void enqueue_cancel_ops_unlocked(socket_type descriptor)
-  {
-    pending_cancellations_.push_back(descriptor);
-  }
-
   // Cancel any operations that are running against the descriptor and remove
   // its registration from the reactor.
-  void close_descriptor(socket_type descriptor)
+  void close_descriptor(socket_type descriptor, per_descriptor_data&)
   {
     asio::detail::mutex::scoped_lock lock(mutex_);
 
@@ -349,16 +403,16 @@ private:
 
     // Dispatch any operation cancellations that were made while the select
     // loop was not running.
-    read_op_queue_.dispatch_cancellations();
-    write_op_queue_.dispatch_cancellations();
-    except_op_queue_.dispatch_cancellations();
+    read_op_queue_.perform_cancellations();
+    write_op_queue_.perform_cancellations();
+    except_op_queue_.perform_cancellations();
     for (std::size_t i = 0; i < timer_queues_.size(); ++i)
       timer_queues_[i]->dispatch_cancellations();
 
     // Check if the thread is supposed to stop.
     if (stop_thread_)
     {
-      cleanup_operations_and_timers(lock);
+      complete_operations_and_timers(lock);
       return;
     }
 
@@ -367,7 +421,7 @@ private:
     if (!block && read_op_queue_.empty() && write_op_queue_.empty()
         && except_op_queue_.empty() && all_timer_queues_are_empty())
     {
-      cleanup_operations_and_timers(lock);
+      complete_operations_and_timers(lock);
       return;
     }
 
@@ -387,7 +441,7 @@ private:
     lock.lock();
     wait_in_progress_ = false;
 
-    // Block signals while dispatching operations.
+    // Block signals while performing operations.
     asio::detail::signal_blocker sb;
 
     // Dispatch the waiting events.
@@ -407,22 +461,22 @@ private:
         {
           asio::error_code error(
               events[i].data, asio::error::get_system_category());
-          except_op_queue_.dispatch_all_operations(descriptor, error);
-          read_op_queue_.dispatch_all_operations(descriptor, error);
+          except_op_queue_.perform_all_operations(descriptor, error);
+          read_op_queue_.perform_all_operations(descriptor, error);
         }
         else if (events[i].flags & EV_OOBAND)
         {
           asio::error_code error;
-          more_except = except_op_queue_.dispatch_operation(descriptor, error);
+          more_except = except_op_queue_.perform_operation(descriptor, error);
           if (events[i].data > 0)
-            more_reads = read_op_queue_.dispatch_operation(descriptor, error);
+            more_reads = read_op_queue_.perform_operation(descriptor, error);
           else
             more_reads = read_op_queue_.has_operation(descriptor);
         }
         else
         {
           asio::error_code error;
-          more_reads = read_op_queue_.dispatch_operation(descriptor, error);
+          more_reads = read_op_queue_.perform_operation(descriptor, error);
           more_except = except_op_queue_.has_operation(descriptor);
         }
 
@@ -438,8 +492,8 @@ private:
         {
           asio::error_code error(errno,
               asio::error::get_system_category());
-          except_op_queue_.dispatch_all_operations(descriptor, error);
-          read_op_queue_.dispatch_all_operations(descriptor, error);
+          except_op_queue_.perform_all_operations(descriptor, error);
+          read_op_queue_.perform_all_operations(descriptor, error);
         }
       }
       else if (events[i].filter == EVFILT_WRITE)
@@ -450,12 +504,12 @@ private:
         {
           asio::error_code error(
               events[i].data, asio::error::get_system_category());
-          write_op_queue_.dispatch_all_operations(descriptor, error);
+          write_op_queue_.perform_all_operations(descriptor, error);
         }
         else
         {
           asio::error_code error;
-          more_writes = write_op_queue_.dispatch_operation(descriptor, error);
+          more_writes = write_op_queue_.perform_operation(descriptor, error);
         }
 
         // Update the descriptor in the kqueue.
@@ -468,14 +522,14 @@ private:
         {
           asio::error_code error(errno,
               asio::error::get_system_category());
-          write_op_queue_.dispatch_all_operations(descriptor, error);
+          write_op_queue_.perform_all_operations(descriptor, error);
         }
       }
     }
 
-    read_op_queue_.dispatch_cancellations();
-    write_op_queue_.dispatch_cancellations();
-    except_op_queue_.dispatch_cancellations();
+    read_op_queue_.perform_cancellations();
+    write_op_queue_.perform_cancellations();
+    except_op_queue_.perform_cancellations();
     for (std::size_t i = 0; i < timer_queues_.size(); ++i)
     {
       timer_queues_[i]->dispatch_timers();
@@ -491,7 +545,7 @@ private:
     need_kqueue_wait_ = !read_op_queue_.empty()
       || !write_op_queue_.empty() || !except_op_queue_.empty();
 
-    cleanup_operations_and_timers(lock);
+    complete_operations_and_timers(lock);
   }
 
   // Run the select loop in the thread.
@@ -592,16 +646,16 @@ private:
   // destructors may make calls back into this reactor. We make a copy of the
   // vector of timer queues since the original may be modified while the lock
   // is not held.
-  void cleanup_operations_and_timers(
+  void complete_operations_and_timers(
       asio::detail::mutex::scoped_lock& lock)
   {
     timer_queues_for_cleanup_ = timer_queues_;
     lock.unlock();
-    read_op_queue_.cleanup_operations();
-    write_op_queue_.cleanup_operations();
-    except_op_queue_.cleanup_operations();
+    read_op_queue_.complete_operations();
+    write_op_queue_.complete_operations();
+    except_op_queue_.complete_operations();
     for (std::size_t i = 0; i < timer_queues_for_cleanup_.size(); ++i)
-      timer_queues_for_cleanup_[i]->cleanup_timers();
+      timer_queues_for_cleanup_[i]->complete_timers();
   }
 
   // Mutex to protect access to internal data.
