@@ -26,6 +26,7 @@
 #include "asio/io_service.hpp"
 #include "asio/socket_base.hpp"
 #include "asio/detail/bind_handler.hpp"
+#include "asio/detail/handler_base_from_member.hpp"
 #include "asio/detail/noncopyable.hpp"
 #include "asio/detail/service_base.hpp"
 #include "asio/detail/socket_holder.hpp"
@@ -72,10 +73,21 @@ public:
 
     enum
     {
-      user_set_non_blocking = 1, // The user wants a non-blocking socket.
-      internal_non_blocking = 2, // The socket has been set non-blocking.
-      enable_connection_aborted = 4, // User wants connection_aborted errors.
-      user_set_linger = 8 // The user set the linger option.
+      // The user wants a non-blocking socket.
+      user_set_non_blocking = 1,
+
+      // The implementation wants a non-blocking socket (in order to be able to
+      // perform asynchronous read and write operations).
+      internal_non_blocking = 2,
+
+      // Helper "flag" used to determine whether the socket is non-blocking.
+      non_blocking = user_set_non_blocking | internal_non_blocking,
+
+      // User wants connection_aborted errors, which are disabled by default.
+      enable_connection_aborted = 4,
+
+      // The user set the linger option. Needs to be checked when closing. 
+      user_set_linger = 8
     };
 
     // Flags indicating the current state of the socket.
@@ -83,6 +95,9 @@ public:
 
     // The protocol associated with the socket.
     protocol_type protocol_;
+
+    // Per-descriptor data used by the reactor.
+    typename Reactor::per_descriptor_data reactor_data_;
   };
 
   // The maximum number of buffers to support in a single operation.
@@ -94,6 +109,7 @@ public:
         reactive_socket_service<Protocol, Reactor> >(io_service),
       reactor_(asio::use_service<Reactor>(io_service))
   {
+    reactor_.init_task();
   }
 
   // Destroy all user-defined handler objects owned by the service.
@@ -113,14 +129,14 @@ public:
   {
     if (impl.socket_ != invalid_socket)
     {
-      reactor_.close_descriptor(impl.socket_);
+      reactor_.close_descriptor(impl.socket_, impl.reactor_data_);
 
-      if (impl.flags_ & implementation_type::internal_non_blocking)
+      if (impl.flags_ & implementation_type::non_blocking)
       {
         ioctl_arg_type non_blocking = 0;
         asio::error_code ignored_ec;
         socket_ops::ioctl(impl.socket_, FIONBIO, &non_blocking, ignored_ec);
-        impl.flags_ &= ~implementation_type::internal_non_blocking;
+        impl.flags_ &= ~implementation_type::non_blocking;
       }
 
       if (impl.flags_ & implementation_type::user_set_linger)
@@ -155,7 +171,7 @@ public:
     if (sock.get() == invalid_socket)
       return ec;
 
-    if (int err = reactor_.register_descriptor(sock.get()))
+    if (int err = reactor_.register_descriptor(sock.get(), impl.reactor_data_))
     {
       ec = asio::error_code(err,
           asio::error::get_system_category());
@@ -180,7 +196,8 @@ public:
       return ec;
     }
 
-    if (int err = reactor_.register_descriptor(native_socket))
+    if (int err = reactor_.register_descriptor(
+          native_socket, impl.reactor_data_))
     {
       ec = asio::error_code(err,
           asio::error::get_system_category());
@@ -206,14 +223,14 @@ public:
   {
     if (is_open(impl))
     {
-      reactor_.close_descriptor(impl.socket_);
+      reactor_.close_descriptor(impl.socket_, impl.reactor_data_);
 
-      if (impl.flags_ & implementation_type::internal_non_blocking)
+      if (impl.flags_ & implementation_type::non_blocking)
       {
         ioctl_arg_type non_blocking = 0;
         asio::error_code ignored_ec;
         socket_ops::ioctl(impl.socket_, FIONBIO, &non_blocking, ignored_ec);
-        impl.flags_ &= ~implementation_type::internal_non_blocking;
+        impl.flags_ &= ~implementation_type::non_blocking;
       }
 
       if (socket_ops::close(impl.socket_, ec) == socket_error_retval)
@@ -242,7 +259,7 @@ public:
       return ec;
     }
 
-    reactor_.cancel_ops(impl.socket_);
+    reactor_.cancel_ops(impl.socket_, impl.reactor_data_);
     ec = asio::error_code();
     return ec;
   }
@@ -427,11 +444,35 @@ public:
 
     if (command.name() == static_cast<int>(FIONBIO))
     {
+      // Flags are manipulated in a temporary variable so that the socket
+      // implementation is not updated unless the ioctl operation succeeds.
+      unsigned char new_flags = impl.flags_;
       if (command.get())
-        impl.flags_ |= implementation_type::user_set_non_blocking;
+        new_flags |= implementation_type::user_set_non_blocking;
       else
-        impl.flags_ &= ~implementation_type::user_set_non_blocking;
-      ec = asio::error_code();
+        new_flags &= ~implementation_type::user_set_non_blocking;
+
+      // Perform ioctl on socket if the non-blocking state has changed.
+      if (!(impl.flags_ & implementation_type::non_blocking)
+          && (new_flags & implementation_type::non_blocking))
+      {
+        ioctl_arg_type non_blocking = 1;
+        socket_ops::ioctl(impl.socket_, FIONBIO, &non_blocking, ec);
+      }
+      else if ((impl.flags_ & implementation_type::non_blocking)
+          && !(new_flags & implementation_type::non_blocking))
+      {
+        ioctl_arg_type non_blocking = 0;
+        socket_ops::ioctl(impl.socket_, FIONBIO, &non_blocking, ec);
+      }
+      else
+      {
+        ec = asio::error_code();
+      }
+
+      // Update socket implementation's flags only if successful.
+      if (!ec)
+        impl.flags_ = new_flags;
     }
     else
     {
@@ -524,18 +565,6 @@ public:
       return 0;
     }
 
-    // Make socket non-blocking if user wants non-blocking.
-    if (impl.flags_ & implementation_type::user_set_non_blocking)
-    {
-      if (!(impl.flags_ & implementation_type::internal_non_blocking))
-      {
-        ioctl_arg_type non_blocking = 1;
-        if (socket_ops::ioctl(impl.socket_, FIONBIO, &non_blocking, ec))
-          return 0;
-        impl.flags_ |= implementation_type::internal_non_blocking;
-      }
-    }
-
     // Send the data.
     for (;;)
     {
@@ -575,27 +604,29 @@ public:
   }
 
   template <typename ConstBufferSequence, typename Handler>
-  class send_handler
+  class send_operation :
+    public handler_base_from_member<Handler>
   {
   public:
-    send_handler(socket_type socket, asio::io_service& io_service,
+    send_operation(socket_type socket, asio::io_service& io_service,
         const ConstBufferSequence& buffers, socket_base::message_flags flags,
         Handler handler)
-      : socket_(socket),
+      : handler_base_from_member<Handler>(handler),
+        socket_(socket),
         io_service_(io_service),
         work_(io_service),
         buffers_(buffers),
-        flags_(flags),
-        handler_(handler)
+        flags_(flags)
     {
     }
 
-    bool operator()(const asio::error_code& result)
+    bool perform(asio::error_code& ec,
+        std::size_t& bytes_transferred)
     {
       // Check whether the operation was successful.
-      if (result)
+      if (ec)
       {
-        io_service_.post(bind_handler(handler_, result, 0));
+        bytes_transferred = 0;
         return true;
       }
 
@@ -613,7 +644,6 @@ public:
       }
 
       // Send the data.
-      asio::error_code ec;
       int bytes = socket_ops::send(socket_, bufs, i, flags_, ec);
 
       // Check if we need to run the operation again.
@@ -621,17 +651,22 @@ public:
           || ec == asio::error::try_again)
         return false;
 
-      io_service_.post(bind_handler(handler_, ec, bytes < 0 ? 0 : bytes));
+      bytes_transferred = (bytes < 0 ? 0 : bytes);
       return true;
     }
 
+    void complete(const asio::error_code& ec,
+        std::size_t bytes_transferred)
+    {
+      io_service_.post(bind_handler(this->handler_, ec, bytes_transferred));
+    }
+
   private:
     socket_type socket_;
     asio::io_service& io_service_;
     asio::io_service::work work_;
     ConstBufferSequence buffers_;
     socket_base::message_flags flags_;
-    Handler handler_;
   };
 
   // Start an asynchronous send. The data being sent must be valid for the
@@ -672,41 +707,52 @@ public:
       // Make socket non-blocking.
       if (!(impl.flags_ & implementation_type::internal_non_blocking))
       {
-        ioctl_arg_type non_blocking = 1;
-        asio::error_code ec;
-        if (socket_ops::ioctl(impl.socket_, FIONBIO, &non_blocking, ec))
+        if (!(impl.flags_ & implementation_type::non_blocking))
         {
-          this->get_io_service().post(bind_handler(handler, ec, 0));
-          return;
+          ioctl_arg_type non_blocking = 1;
+          asio::error_code ec;
+          if (socket_ops::ioctl(impl.socket_, FIONBIO, &non_blocking, ec))
+          {
+            this->get_io_service().post(bind_handler(handler, ec, 0));
+            return;
+          }
         }
         impl.flags_ |= implementation_type::internal_non_blocking;
       }
 
-      reactor_.start_write_op(impl.socket_,
-          send_handler<ConstBufferSequence, Handler>(
+      reactor_.start_write_op(impl.socket_, impl.reactor_data_,
+          send_operation<ConstBufferSequence, Handler>(
             impl.socket_, this->get_io_service(), buffers, flags, handler));
     }
   }
 
   template <typename Handler>
-  class null_buffers_handler
+  class null_buffers_operation :
+    public handler_base_from_member<Handler>
   {
   public:
-    null_buffers_handler(asio::io_service& io_service, Handler handler)
-      : work_(io_service),
-        handler_(handler)
+    null_buffers_operation(asio::io_service& io_service, Handler handler)
+      : handler_base_from_member<Handler>(handler),
+        work_(io_service)
     {
     }
 
-    bool operator()(const asio::error_code& result)
+    bool perform(asio::error_code&,
+        std::size_t& bytes_transferred)
     {
-      work_.get_io_service().post(bind_handler(handler_, result, 0));
+      bytes_transferred = 0;
       return true;
     }
 
+    void complete(const asio::error_code& ec,
+        std::size_t bytes_transferred)
+    {
+      work_.get_io_service().post(bind_handler(
+            this->handler_, ec, bytes_transferred));
+    }
+
   private:
     asio::io_service::work work_;
-    Handler handler_;
   };
 
   // Start an asynchronous wait until data can be sent without blocking.
@@ -721,8 +767,8 @@ public:
     }
     else
     {
-      reactor_.start_write_op(impl.socket_,
-          null_buffers_handler<Handler>(this->get_io_service(), handler),
+      reactor_.start_write_op(impl.socket_, impl.reactor_data_,
+          null_buffers_operation<Handler>(this->get_io_service(), handler),
           false);
     }
   }
@@ -753,18 +799,6 @@ public:
           asio::buffer_size(buffer));
     }
 
-    // Make socket non-blocking if user wants non-blocking.
-    if (impl.flags_ & implementation_type::user_set_non_blocking)
-    {
-      if (!(impl.flags_ & implementation_type::internal_non_blocking))
-      {
-        ioctl_arg_type non_blocking = 1;
-        if (socket_ops::ioctl(impl.socket_, FIONBIO, &non_blocking, ec))
-          return 0;
-        impl.flags_ |= implementation_type::internal_non_blocking;
-      }
-    }
-
     // Send the data.
     for (;;)
     {
@@ -806,28 +840,30 @@ public:
   }
 
   template <typename ConstBufferSequence, typename Handler>
-  class send_to_handler
+  class send_to_operation :
+    public handler_base_from_member<Handler>
   {
   public:
-    send_to_handler(socket_type socket, asio::io_service& io_service,
+    send_to_operation(socket_type socket, asio::io_service& io_service,
         const ConstBufferSequence& buffers, const endpoint_type& endpoint,
         socket_base::message_flags flags, Handler handler)
-      : socket_(socket),
+      : handler_base_from_member<Handler>(handler),
+        socket_(socket),
         io_service_(io_service),
         work_(io_service),
         buffers_(buffers),
         destination_(endpoint),
-        flags_(flags),
-        handler_(handler)
+        flags_(flags)
     {
     }
 
-    bool operator()(const asio::error_code& result)
+    bool perform(asio::error_code& ec,
+        std::size_t& bytes_transferred)
     {
       // Check whether the operation was successful.
-      if (result)
+      if (ec)
       {
-        io_service_.post(bind_handler(handler_, result, 0));
+        bytes_transferred = 0;
         return true;
       }
 
@@ -845,7 +881,6 @@ public:
       }
 
       // Send the data.
-      asio::error_code ec;
       int bytes = socket_ops::sendto(socket_, bufs, i, flags_,
           destination_.data(), destination_.size(), ec);
 
@@ -854,10 +889,16 @@ public:
           || ec == asio::error::try_again)
         return false;
 
-      io_service_.post(bind_handler(handler_, ec, bytes < 0 ? 0 : bytes));
+      bytes_transferred = (bytes < 0 ? 0 : bytes);
       return true;
     }
 
+    void complete(const asio::error_code& ec,
+        std::size_t bytes_transferred)
+    {
+      io_service_.post(bind_handler(this->handler_, ec, bytes_transferred));
+    }
+
   private:
     socket_type socket_;
     asio::io_service& io_service_;
@@ -865,7 +906,6 @@ public:
     ConstBufferSequence buffers_;
     endpoint_type destination_;
     socket_base::message_flags flags_;
-    Handler handler_;
   };
 
   // Start an asynchronous send. The data being sent must be valid for the
@@ -886,18 +926,21 @@ public:
       // Make socket non-blocking.
       if (!(impl.flags_ & implementation_type::internal_non_blocking))
       {
-        ioctl_arg_type non_blocking = 1;
-        asio::error_code ec;
-        if (socket_ops::ioctl(impl.socket_, FIONBIO, &non_blocking, ec))
+        if (!(impl.flags_ & implementation_type::non_blocking))
         {
-          this->get_io_service().post(bind_handler(handler, ec, 0));
-          return;
+          ioctl_arg_type non_blocking = 1;
+          asio::error_code ec;
+          if (socket_ops::ioctl(impl.socket_, FIONBIO, &non_blocking, ec))
+          {
+            this->get_io_service().post(bind_handler(handler, ec, 0));
+            return;
+          }
         }
         impl.flags_ |= implementation_type::internal_non_blocking;
       }
 
-      reactor_.start_write_op(impl.socket_,
-          send_to_handler<ConstBufferSequence, Handler>(
+      reactor_.start_write_op(impl.socket_, impl.reactor_data_,
+          send_to_operation<ConstBufferSequence, Handler>(
             impl.socket_, this->get_io_service(), buffers,
             destination, flags, handler));
     }
@@ -915,8 +958,8 @@ public:
     }
     else
     {
-      reactor_.start_write_op(impl.socket_,
-          null_buffers_handler<Handler>(this->get_io_service(), handler),
+      reactor_.start_write_op(impl.socket_, impl.reactor_data_,
+          null_buffers_operation<Handler>(this->get_io_service(), handler),
           false);
     }
   }
@@ -955,18 +998,6 @@ public:
       return 0;
     }
 
-    // Make socket non-blocking if user wants non-blocking.
-    if (impl.flags_ & implementation_type::user_set_non_blocking)
-    {
-      if (!(impl.flags_ & implementation_type::internal_non_blocking))
-      {
-        ioctl_arg_type non_blocking = 1;
-        if (socket_ops::ioctl(impl.socket_, FIONBIO, &non_blocking, ec))
-          return 0;
-        impl.flags_ |= implementation_type::internal_non_blocking;
-      }
-    }
-
     // Receive some data.
     for (;;)
     {
@@ -978,7 +1009,7 @@ public:
         return bytes_recvd;
 
       // Check for EOF.
-      if (bytes_recvd == 0)
+      if (bytes_recvd == 0 && impl.protocol_.type() == SOCK_STREAM)
       {
         ec = asio::error::eof;
         return 0;
@@ -997,8 +1028,7 @@ public:
   }
 
   // Wait until data can be received without blocking.
-  size_t receive(implementation_type& impl,
-      const null_buffers& buffers,
+  size_t receive(implementation_type& impl, const null_buffers&,
       socket_base::message_flags, asio::error_code& ec)
   {
     if (!is_open(impl))
@@ -1014,27 +1044,31 @@ public:
   }
 
   template <typename MutableBufferSequence, typename Handler>
-  class receive_handler
+  class receive_operation :
+    public handler_base_from_member<Handler>
   {
   public:
-    receive_handler(socket_type socket, asio::io_service& io_service,
-        const MutableBufferSequence& buffers, socket_base::message_flags flags,
-        Handler handler)
-      : socket_(socket),
+    receive_operation(socket_type socket, int protocol_type,
+        asio::io_service& io_service,
+        const MutableBufferSequence& buffers,
+        socket_base::message_flags flags, Handler handler)
+      : handler_base_from_member<Handler>(handler),
+        socket_(socket),
+        protocol_type_(protocol_type),
         io_service_(io_service),
         work_(io_service),
         buffers_(buffers),
-        flags_(flags),
-        handler_(handler)
+        flags_(flags)
     {
     }
 
-    bool operator()(const asio::error_code& result)
+    bool perform(asio::error_code& ec,
+        std::size_t& bytes_transferred)
     {
       // Check whether the operation was successful.
-      if (result)
+      if (ec)
       {
-        io_service_.post(bind_handler(handler_, result, 0));
+        bytes_transferred = 0;
         return true;
       }
 
@@ -1052,9 +1086,8 @@ public:
       }
 
       // Receive some data.
-      asio::error_code ec;
       int bytes = socket_ops::recv(socket_, bufs, i, flags_, ec);
-      if (bytes == 0)
+      if (bytes == 0 && protocol_type_ == SOCK_STREAM)
         ec = asio::error::eof;
 
       // Check if we need to run the operation again.
@@ -1062,17 +1095,23 @@ public:
           || ec == asio::error::try_again)
         return false;
 
-      io_service_.post(bind_handler(handler_, ec, bytes < 0 ? 0 : bytes));
+      bytes_transferred = (bytes < 0 ? 0 : bytes);
       return true;
     }
 
+    void complete(const asio::error_code& ec,
+        std::size_t bytes_transferred)
+    {
+      io_service_.post(bind_handler(this->handler_, ec, bytes_transferred));
+    }
+
   private:
     socket_type socket_;
+    int protocol_type_;
     asio::io_service& io_service_;
     asio::io_service::work work_;
     MutableBufferSequence buffers_;
     socket_base::message_flags flags_;
-    Handler handler_;
   };
 
   // Start an asynchronous receive. The buffer for the data being received
@@ -1114,27 +1153,32 @@ public:
       // Make socket non-blocking.
       if (!(impl.flags_ & implementation_type::internal_non_blocking))
       {
-        ioctl_arg_type non_blocking = 1;
-        asio::error_code ec;
-        if (socket_ops::ioctl(impl.socket_, FIONBIO, &non_blocking, ec))
+        if (!(impl.flags_ & implementation_type::non_blocking))
         {
-          this->get_io_service().post(bind_handler(handler, ec, 0));
-          return;
+          ioctl_arg_type non_blocking = 1;
+          asio::error_code ec;
+          if (socket_ops::ioctl(impl.socket_, FIONBIO, &non_blocking, ec))
+          {
+            this->get_io_service().post(bind_handler(handler, ec, 0));
+            return;
+          }
         }
         impl.flags_ |= implementation_type::internal_non_blocking;
       }
 
       if (flags & socket_base::message_out_of_band)
       {
-        reactor_.start_except_op(impl.socket_,
-            receive_handler<MutableBufferSequence, Handler>(
-              impl.socket_, this->get_io_service(), buffers, flags, handler));
+        reactor_.start_except_op(impl.socket_, impl.reactor_data_,
+            receive_operation<MutableBufferSequence, Handler>(
+              impl.socket_, impl.protocol_.type(),
+              this->get_io_service(), buffers, flags, handler));
       }
       else
       {
-        reactor_.start_read_op(impl.socket_,
-            receive_handler<MutableBufferSequence, Handler>(
-              impl.socket_, this->get_io_service(), buffers, flags, handler));
+        reactor_.start_read_op(impl.socket_, impl.reactor_data_,
+            receive_operation<MutableBufferSequence, Handler>(
+              impl.socket_, impl.protocol_.type(),
+              this->get_io_service(), buffers, flags, handler));
       }
     }
   }
@@ -1151,13 +1195,13 @@ public:
     }
     else if (flags & socket_base::message_out_of_band)
     {
-      reactor_.start_except_op(impl.socket_,
-          null_buffers_handler<Handler>(this->get_io_service(), handler));
+      reactor_.start_except_op(impl.socket_, impl.reactor_data_,
+          null_buffers_operation<Handler>(this->get_io_service(), handler));
     }
     else
     {
-      reactor_.start_read_op(impl.socket_,
-          null_buffers_handler<Handler>(this->get_io_service(), handler),
+      reactor_.start_read_op(impl.socket_, impl.reactor_data_,
+          null_buffers_operation<Handler>(this->get_io_service(), handler),
           false);
     }
   }
@@ -1189,18 +1233,6 @@ public:
           asio::buffer_size(buffer));
     }
 
-    // Make socket non-blocking if user wants non-blocking.
-    if (impl.flags_ & implementation_type::user_set_non_blocking)
-    {
-      if (!(impl.flags_ & implementation_type::internal_non_blocking))
-      {
-        ioctl_arg_type non_blocking = 1;
-        if (socket_ops::ioctl(impl.socket_, FIONBIO, &non_blocking, ec))
-          return 0;
-        impl.flags_ |= implementation_type::internal_non_blocking;
-      }
-    }
-
     // Receive some data.
     for (;;)
     {
@@ -1217,7 +1249,7 @@ public:
       }
 
       // Check for EOF.
-      if (bytes_recvd == 0)
+      if (bytes_recvd == 0 && impl.protocol_.type() == SOCK_STREAM)
       {
         ec = asio::error::eof;
         return 0;
@@ -1236,9 +1268,9 @@ public:
   }
 
   // Wait until data can be received without blocking.
-  size_t receive_from(implementation_type& impl,
-      const null_buffers& buffers, endpoint_type& sender_endpoint,
-      socket_base::message_flags, asio::error_code& ec)
+  size_t receive_from(implementation_type& impl, const null_buffers&,
+      endpoint_type& sender_endpoint, socket_base::message_flags,
+      asio::error_code& ec)
   {
     if (!is_open(impl))
     {
@@ -1256,29 +1288,32 @@ public:
   }
 
   template <typename MutableBufferSequence, typename Handler>
-  class receive_from_handler
+  class receive_from_operation :
+    public handler_base_from_member<Handler>
   {
   public:
-    receive_from_handler(socket_type socket,
+    receive_from_operation(socket_type socket, int protocol_type,
         asio::io_service& io_service,
         const MutableBufferSequence& buffers, endpoint_type& endpoint,
         socket_base::message_flags flags, Handler handler)
-      : socket_(socket),
+      : handler_base_from_member<Handler>(handler),
+        socket_(socket),
+        protocol_type_(protocol_type),
         io_service_(io_service),
         work_(io_service),
         buffers_(buffers),
         sender_endpoint_(endpoint),
-        flags_(flags),
-        handler_(handler)
+        flags_(flags)
     {
     }
 
-    bool operator()(const asio::error_code& result)
+    bool perform(asio::error_code& ec,
+        std::size_t& bytes_transferred)
     {
       // Check whether the operation was successful.
-      if (result)
+      if (ec)
       {
-        io_service_.post(bind_handler(handler_, result, 0));
+        bytes_transferred = 0;
         return true;
       }
 
@@ -1297,10 +1332,9 @@ public:
 
       // Receive some data.
       std::size_t addr_len = sender_endpoint_.capacity();
-      asio::error_code ec;
       int bytes = socket_ops::recvfrom(socket_, bufs, i, flags_,
           sender_endpoint_.data(), &addr_len, ec);
-      if (bytes == 0)
+      if (bytes == 0 && protocol_type_ == SOCK_STREAM)
         ec = asio::error::eof;
 
       // Check if we need to run the operation again.
@@ -1309,18 +1343,24 @@ public:
         return false;
 
       sender_endpoint_.resize(addr_len);
-      io_service_.post(bind_handler(handler_, ec, bytes < 0 ? 0 : bytes));
+      bytes_transferred = (bytes < 0 ? 0 : bytes);
       return true;
     }
 
+    void complete(const asio::error_code& ec,
+        std::size_t bytes_transferred)
+    {
+      io_service_.post(bind_handler(this->handler_, ec, bytes_transferred));
+    }
+
   private:
     socket_type socket_;
+    int protocol_type_;
     asio::io_service& io_service_;
     asio::io_service::work work_;
     MutableBufferSequence buffers_;
     endpoint_type& sender_endpoint_;
     socket_base::message_flags flags_;
-    Handler handler_;
   };
 
   // Start an asynchronous receive. The buffer for the data being received and
@@ -1341,20 +1381,23 @@ public:
       // Make socket non-blocking.
       if (!(impl.flags_ & implementation_type::internal_non_blocking))
       {
-        ioctl_arg_type non_blocking = 1;
-        asio::error_code ec;
-        if (socket_ops::ioctl(impl.socket_, FIONBIO, &non_blocking, ec))
+        if (!(impl.flags_ & implementation_type::non_blocking))
         {
-          this->get_io_service().post(bind_handler(handler, ec, 0));
-          return;
+          ioctl_arg_type non_blocking = 1;
+          asio::error_code ec;
+          if (socket_ops::ioctl(impl.socket_, FIONBIO, &non_blocking, ec))
+          {
+            this->get_io_service().post(bind_handler(handler, ec, 0));
+            return;
+          }
         }
         impl.flags_ |= implementation_type::internal_non_blocking;
       }
 
-      reactor_.start_read_op(impl.socket_,
-          receive_from_handler<MutableBufferSequence, Handler>(
-            impl.socket_, this->get_io_service(), buffers,
-            sender_endpoint, flags, handler));
+      reactor_.start_read_op(impl.socket_, impl.reactor_data_,
+          receive_from_operation<MutableBufferSequence, Handler>(
+            impl.socket_, impl.protocol_.type(), this->get_io_service(),
+            buffers, sender_endpoint, flags, handler));
     }
   }
 
@@ -1376,13 +1419,13 @@ public:
 
       if (flags & socket_base::message_out_of_band)
       {
-        reactor_.start_except_op(impl.socket_,
-            null_buffers_handler<Handler>(this->get_io_service(), handler));
+        reactor_.start_except_op(impl.socket_, impl.reactor_data_,
+            null_buffers_operation<Handler>(this->get_io_service(), handler));
       }
       else
       {
-        reactor_.start_read_op(impl.socket_,
-            null_buffers_handler<Handler>(this->get_io_service(), handler),
+        reactor_.start_read_op(impl.socket_, impl.reactor_data_,
+            null_buffers_operation<Handler>(this->get_io_service(), handler),
             false);
       }
     }
@@ -1406,18 +1449,6 @@ public:
       return ec;
     }
 
-    // Make socket non-blocking if user wants non-blocking.
-    if (impl.flags_ & implementation_type::user_set_non_blocking)
-    {
-      if (!(impl.flags_ & implementation_type::internal_non_blocking))
-      {
-        ioctl_arg_type non_blocking = 1;
-        if (socket_ops::ioctl(impl.socket_, FIONBIO, &non_blocking, ec))
-          return ec;
-        impl.flags_ |= implementation_type::internal_non_blocking;
-      }
-    }
-
     // Accept a socket.
     for (;;)
     {
@@ -1479,35 +1510,32 @@ public:
   }
 
   template <typename Socket, typename Handler>
-  class accept_handler
+  class accept_operation :
+    public handler_base_from_member<Handler>
   {
   public:
-    accept_handler(socket_type socket, asio::io_service& io_service,
+    accept_operation(socket_type socket, asio::io_service& io_service,
         Socket& peer, const protocol_type& protocol,
         endpoint_type* peer_endpoint, bool enable_connection_aborted,
         Handler handler)
-      : socket_(socket),
+      : handler_base_from_member<Handler>(handler),
+        socket_(socket),
         io_service_(io_service),
         work_(io_service),
         peer_(peer),
         protocol_(protocol),
         peer_endpoint_(peer_endpoint),
-        enable_connection_aborted_(enable_connection_aborted),
-        handler_(handler)
+        enable_connection_aborted_(enable_connection_aborted)
     {
     }
 
-    bool operator()(const asio::error_code& result)
+    bool perform(asio::error_code& ec, std::size_t&)
     {
       // Check whether the operation was successful.
-      if (result)
-      {
-        io_service_.post(bind_handler(handler_, result));
+      if (ec)
         return true;
-      }
 
       // Accept the waiting connection.
-      asio::error_code ec;
       socket_holder new_socket;
       std::size_t addr_len = 0;
       if (peer_endpoint_)
@@ -1543,10 +1571,14 @@ public:
           new_socket.release();
       }
 
-      io_service_.post(bind_handler(handler_, ec));
       return true;
     }
 
+    void complete(const asio::error_code& ec, std::size_t)
+    {
+      io_service_.post(bind_handler(this->handler_, ec));
+    }
+
   private:
     socket_type socket_;
     asio::io_service& io_service_;
@@ -1555,7 +1587,6 @@ public:
     protocol_type protocol_;
     endpoint_type* peer_endpoint_;
     bool enable_connection_aborted_;
-    Handler handler_;
   };
 
   // Start an asynchronous accept. The peer and peer_endpoint objects
@@ -1579,18 +1610,21 @@ public:
       // Make socket non-blocking.
       if (!(impl.flags_ & implementation_type::internal_non_blocking))
       {
-        ioctl_arg_type non_blocking = 1;
-        asio::error_code ec;
-        if (socket_ops::ioctl(impl.socket_, FIONBIO, &non_blocking, ec))
+        if (!(impl.flags_ & implementation_type::non_blocking))
         {
-          this->get_io_service().post(bind_handler(handler, ec));
-          return;
+          ioctl_arg_type non_blocking = 1;
+          asio::error_code ec;
+          if (socket_ops::ioctl(impl.socket_, FIONBIO, &non_blocking, ec))
+          {
+            this->get_io_service().post(bind_handler(handler, ec));
+            return;
+          }
         }
         impl.flags_ |= implementation_type::internal_non_blocking;
       }
 
-      reactor_.start_read_op(impl.socket_,
-          accept_handler<Socket, Handler>(
+      reactor_.start_read_op(impl.socket_, impl.reactor_data_,
+          accept_operation<Socket, Handler>(
             impl.socket_, this->get_io_service(),
             peer, impl.protocol_, peer_endpoint,
             (impl.flags_ & implementation_type::enable_connection_aborted) != 0,
@@ -1608,86 +1642,80 @@ public:
       return ec;
     }
 
-    if (impl.flags_ & implementation_type::internal_non_blocking)
-    {
-      // Mark the socket as blocking while we perform the connect.
-      ioctl_arg_type non_blocking = 0;
-      if (socket_ops::ioctl(impl.socket_, FIONBIO, &non_blocking, ec))
-        return ec;
-      impl.flags_ &= ~implementation_type::internal_non_blocking;
-    }
-
     // Perform the connect operation.
     socket_ops::connect(impl.socket_,
         peer_endpoint.data(), peer_endpoint.size(), ec);
+    if (ec != asio::error::in_progress
+        && ec != asio::error::would_block)
+    {
+      // The connect operation finished immediately.
+      return ec;
+    }
+
+    // Wait for socket to become ready.
+    if (socket_ops::poll_connect(impl.socket_, ec) < 0)
+      return ec;
+
+    // Get the error code from the connect operation.
+    int connect_error = 0;
+    size_t connect_error_len = sizeof(connect_error);
+    if (socket_ops::getsockopt(impl.socket_, SOL_SOCKET, SO_ERROR,
+          &connect_error, &connect_error_len, ec) == socket_error_retval)
+      return ec;
+
+    // Return the result of the connect operation.
+    ec = asio::error_code(connect_error,
+        asio::error::get_system_category());
     return ec;
   }
 
   template <typename Handler>
-  class connect_handler
+  class connect_operation :
+    public handler_base_from_member<Handler>
   {
   public:
-    connect_handler(socket_type socket, boost::shared_ptr<bool> completed,
-        asio::io_service& io_service, Reactor& reactor, Handler handler)
-      : socket_(socket),
-        completed_(completed),
+    connect_operation(socket_type socket,
+        asio::io_service& io_service, Handler handler)
+      : handler_base_from_member<Handler>(handler),
+        socket_(socket),
         io_service_(io_service),
-        work_(io_service),
-        reactor_(reactor),
-        handler_(handler)
+        work_(io_service)
     {
     }
 
-    bool operator()(const asio::error_code& result)
+    bool perform(asio::error_code& ec, std::size_t&)
     {
-      // Check whether a handler has already been called for the connection.
-      // If it has, then we don't want to do anything in this handler.
-      if (*completed_)
-        return true;
-
-      // Cancel the other reactor operation for the connection.
-      *completed_ = true;
-      reactor_.enqueue_cancel_ops_unlocked(socket_);
-
       // Check whether the operation was successful.
-      if (result)
-      {
-        io_service_.post(bind_handler(handler_, result));
+      if (ec)
         return true;
-      }
 
       // Get the error code from the connect operation.
       int connect_error = 0;
       size_t connect_error_len = sizeof(connect_error);
-      asio::error_code ec;
       if (socket_ops::getsockopt(socket_, SOL_SOCKET, SO_ERROR,
             &connect_error, &connect_error_len, ec) == socket_error_retval)
-      {
-        io_service_.post(bind_handler(handler_, ec));
         return true;
-      }
 
-      // If connection failed then post the handler with the error code.
+      // The connection failed so the handler will be posted with an error code.
       if (connect_error)
       {
         ec = asio::error_code(connect_error,
             asio::error::get_system_category());
-        io_service_.post(bind_handler(handler_, ec));
         return true;
       }
 
-      // Post the result of the successful connection operation.
-      io_service_.post(bind_handler(handler_, ec));
       return true;
     }
 
+    void complete(const asio::error_code& ec, std::size_t)
+    {
+      io_service_.post(bind_handler(this->handler_, ec));
+    }
+
   private:
     socket_type socket_;
-    boost::shared_ptr<bool> completed_;
     asio::io_service& io_service_;
     asio::io_service::work work_;
-    Reactor& reactor_;
-    Handler handler_;
   };
 
   // Start an asynchronous connect.
@@ -1705,12 +1733,15 @@ public:
     // Make socket non-blocking.
     if (!(impl.flags_ & implementation_type::internal_non_blocking))
     {
-      ioctl_arg_type non_blocking = 1;
-      asio::error_code ec;
-      if (socket_ops::ioctl(impl.socket_, FIONBIO, &non_blocking, ec))
+      if (!(impl.flags_ & implementation_type::non_blocking))
       {
-        this->get_io_service().post(bind_handler(handler, ec));
-        return;
+        ioctl_arg_type non_blocking = 1;
+        asio::error_code ec;
+        if (socket_ops::ioctl(impl.socket_, FIONBIO, &non_blocking, ec))
+        {
+          this->get_io_service().post(bind_handler(handler, ec));
+          return;
+        }
       }
       impl.flags_ |= implementation_type::internal_non_blocking;
     }
@@ -1731,10 +1762,9 @@ public:
     {
       // The connection is happening in the background, and we need to wait
       // until the socket becomes writeable.
-      boost::shared_ptr<bool> completed(new bool(false));
-      reactor_.start_write_and_except_ops(impl.socket_,
-          connect_handler<Handler>(impl.socket_, completed,
-            this->get_io_service(), reactor_, handler));
+      reactor_.start_connect_op(impl.socket_, impl.reactor_data_,
+          connect_operation<Handler>(impl.socket_,
+            this->get_io_service(), handler));
     }
     else
     {
